#include "tree_structurer.hpp"

#include <algorithm>
#include <filesystem>
#include <fstream>
#include <sstream>
#include <stdexcept>
#include <functional>
#include <iostream>

namespace fs = std::filesystem;

bool TreeStructurer::should_ignore_dir(const std::string& dirname) {
    static const std::vector<std::string> ignore_list = {
        "build", "venv", "myenv", "dist", "node_modules", "CMakeFiles",
        ".git", ".idea", ".vscode", "__pycache__", "**pycache**"
    };

    // Allow these Python files even if their names start with underscores.
    if (!dirname.empty() && (dirname == "__main__.py" || dirname == "__init__.py")) {
        return false;
    }

    if (std::find(ignore_list.begin(), ignore_list.end(), dirname) != ignore_list.end()) {
        return true;
    }

    if (!dirname.empty()) {
        if (dirname[0] == '.' || dirname[0] == '_') {
            return true;
        }
        if (dirname.find("__") == 0 && dirname.find("__", 2) != std::string::npos) {
            return true;
        }
    }

    return false;
}

bool TreeStructurer::should_ignore_file(const std::string& filename) {
    static const std::vector<std::string> ignore_extensions = {
        ".pyc", ".pyo", ".pyd", ".so", ".dll", ".dylib",
        ".o", ".obj", ".a", ".lib"
    };

    if (!filename.empty() && (filename == "__main__.py" || filename == "__init__.py")) {
        return false;
    }

    if (!filename.empty() && (filename[0] == '.' || filename[0] == '_')) {
        return true;
    }

    fs::path path(filename);
    std::string ext = path.extension().string();
    if (std::find(ignore_extensions.begin(), ignore_extensions.end(), ext) != ignore_extensions.end()) {
        return true;
    }

    return false;
}

std::string TreeStructurer::get_relative_path(const fs::path& path, const fs::path& base) {
    fs::path rel = fs::relative(path, base);
    return rel.string();
}

std::vector<fs::path> TreeStructurer::get_filtered_paths(const fs::path& start) {
    std::vector<fs::path> paths;
    fs::directory_options options = fs::directory_options::skip_permission_denied;

    try {
        if (!fs::exists(start)) {
            throw std::runtime_error("Directory does not exist: " + start.string());
        }
        
        if (!fs::is_directory(start)) {
            throw std::runtime_error("Path is not a directory: " + start.string());
        }

        paths.push_back(start);

        bool is_empty = fs::directory_iterator(start) == fs::directory_iterator();
        if (is_empty) {
            throw std::runtime_error("Directory is empty: " + start.string());
        }

        for (const auto& entry : fs::recursive_directory_iterator(start, options)) {
            const auto& path = entry.path();

            bool should_skip = false;
            for (const auto& component : path) {
                if (should_ignore_dir(component.string())) {
                    should_skip = true;
                    break;
                }
            }
            if (should_skip) continue;

            if (entry.is_directory()) {
                if (!should_ignore_dir(path.filename().string())) {
                    paths.push_back(path);
                }
            } else {
                if (!should_ignore_file(path.filename().string())) {
                    paths.push_back(path);
                }
            }
        }
    } catch (const fs::filesystem_error& e) {
        throw std::runtime_error("Error accessing path: " + std::string(e.what()));
    }

    std::sort(paths.begin(), paths.end());
    return paths;
}

// -----------------------------------------------------------------------------
// Directory Structure Generation (tree printing)
// -----------------------------------------------------------------------------

std::vector<std::string> TreeStructurer::get_directory_structure(const std::string& startpath) {
    std::vector<std::string> lines;
    fs::path start(startpath);
    if (!fs::exists(start)) {
        throw std::runtime_error("Path does not exist: " + startpath);
    }

    // Recursive lambda that traverses the filesystem.
    std::function<void(const fs::path&, const std::string&, bool)> traverse;
    traverse = [&](const fs::path& path, const std::string& indent, bool isLast) {
        std::string line;
        if (indent.empty()) {
            // For the root we simply output the name (appending a directory marker if needed)
            line = path.filename().string();
        } else {
            // If not at the root, prepend a branch marker.
            std::string branch = isLast ?
                (std::string(1, static_cast<char>(TREE_CORNER)) + "── ") :
                (std::string(1, static_cast<char>(TREE_BRANCH)) + "── ");
            line = indent + branch + path.filename().string();
        }
        if (fs::is_directory(path)) {
            line.push_back(static_cast<char>(DIRECTORY_MARKER));
        }
        lines.push_back(line);

        if (fs::is_directory(path)) {
            // Collect and sort the children.
            std::vector<fs::path> children;
            for (auto& entry : fs::directory_iterator(path)) {
                std::string filename = entry.path().filename().string();
                if (entry.is_directory() && should_ignore_dir(filename))
                    continue;
                if (!entry.is_directory() && should_ignore_file(filename))
                    continue;
                children.push_back(entry.path());
            }
            std::sort(children.begin(), children.end(), [](const fs::path& a,
                                                            const fs::path& b) {
                return a.filename().string() < b.filename().string();
            });

            // Recurse into children.
            for (size_t i = 0; i < children.size(); i++) {
                bool last = (i == children.size() - 1);
                // New indent: if we are not at the root then extend the current indent.
                std::string newIndent = indent;
                if (!indent.empty()) {
                    newIndent += isLast ? "    " : u8"│   ";
                }
                traverse(children[i], newIndent, last);
            }
        }
    };

    // Start the traversal with an empty indent.
    traverse(start, "", true);
    return lines;
}

// -----------------------------------------------------------------------------
// Structure Creation from a Tree-like File or String
// -----------------------------------------------------------------------------

void TreeStructurer::create_structure_from_file(const std::string& filepath,
                                                const std::string& target_path) {
    std::vector<std::string> lines = read_structure_file(filepath);
    validate_structure(lines);
    TreeNode root = build_tree_from_lines(lines);
    create_node(root, fs::path(target_path));
}

void TreeStructurer::create_structure_from_string(const std::string& structure,
                                                  const std::string& target_path) {
    std::vector<std::string> lines;
    std::istringstream iss(structure);
    std::string line;
    while (std::getline(iss, line)) {
        if (!line.empty()) {
            lines.push_back(line);
        }
    }
    validate_structure(lines);
    TreeNode root = build_tree_from_lines(lines);
    create_node(root, fs::path(target_path));
}

// -----------------------------------------------------------------------------
// Private Helper Functions (no duplicates)
// -----------------------------------------------------------------------------

// Returns a string consisting of (level * 4) spaces.
std::string TreeStructurer::create_indent(size_t level) {
    return std::string(level * 4, ' ');
}

// Determines the "indent level" of a line by scanning it in 4‑character groups.
// Recognizes either a blank indent ("    " or "│   ") or a branch marker ("├── " or "└── ").
size_t TreeStructurer::get_indent_level(const std::string& line) {
    size_t indent = 0;
    size_t pos = 0;
    while (pos + 4 <= line.size()) {
        std::string group = line.substr(pos, 4);
        if (group == "    " || group == u8"│   ") {
            indent++;
            pos += 4;
            continue;
        }
        if (group == u8"├── " || group == u8"└── ") {
            indent++;
            break;
        }
        break;
    }
    return indent;
}

// Parses a single line of the structure (after knowing its indent level) and returns a TreeNode.
// The function "consumes" indent groups until the branch marker.
TreeStructurer::TreeNode TreeStructurer::parse_structure_line(const std::string& line,
                                                               size_t indent_level) {
    size_t pos = 0;
    size_t groups = 0;
    while (pos + 4 <= line.size() && groups < indent_level) {
        std::string group = line.substr(pos, 4);
        if (group == "    " || group == u8"│   ") {
            pos += 4;
            groups++;
            continue;
        }
        if (group == u8"├── " || group == u8"└── ") {
            pos += 4;
            groups++;
            break;
        }
        break;
    }
    std::string name = line.substr(pos);
    name = sanitize_path(name);
    bool is_file = true;
    if (!name.empty() && name.back() == static_cast<char>(DIRECTORY_MARKER)) {
        is_file = false;
        name.pop_back(); // Remove the directory marker.
    }
    return TreeNode{name, is_file, {}};
}

// Builds a tree (with TreeNode nodes) from the vector of structure lines.
// The first line is assumed to be the root.
TreeStructurer::TreeNode TreeStructurer::build_tree_from_lines(const std::vector<std::string>& lines) {
    if (lines.empty()) {
        throw std::runtime_error("Empty structure provided");
    }
    // Process the first line as the root.
    TreeNode root = parse_structure_line(lines[0], 0);
    if (root.is_file) {
        throw std::runtime_error("Root must be a directory");
    }
    std::vector<TreeNode*> stack;
    stack.push_back(&root);

    // Process each subsequent line.
    for (size_t i = 1; i < lines.size(); ++i) {
        size_t indent = get_indent_level(lines[i]);
        TreeNode node = parse_structure_line(lines[i], indent);
        if (indent > stack.size()) {
            throw std::runtime_error("Invalid indentation structure in the file");
        }
        while (stack.size() > indent) {
            stack.pop_back();
        }
        if (stack.empty()) {
            throw std::runtime_error("Invalid indentation structure in the file");
        }
        stack.back()->children.push_back(node);
        if (!node.is_file) {
            // Push a pointer to the newly added child.
            stack.push_back(&stack.back()->children.back());
        }
    }
    return root;
}

// Recursively creates directories and files on disk according to the tree.
void TreeStructurer::create_node(const TreeNode& node, const fs::path& current_path) {
    fs::path new_path = current_path / node.name;
    try {
        if (node.is_file) {
            // Ensure the parent directory exists.
            fs::path parent = new_path.parent_path();
            if (!fs::exists(parent)) {
                fs::create_directories(parent);
            }
            create_file(new_path);
        } else {
            create_directory(new_path);
            for (const auto& child : node.children) {
                create_node(child, new_path);
            }
        }
    } catch (const fs::filesystem_error& e) {
        throw std::runtime_error("Failed to create path '" + new_path.string() + "': " + e.what());
    }
}

// Returns true if the given line’s last character is a directory marker.
bool TreeStructurer::is_directory_marker(const std::string& line) {
    return (!line.empty() && line.back() == static_cast<char>(DIRECTORY_MARKER));
}

// Creates a directory (and any necessary parent directories).
void TreeStructurer::create_directory(const fs::path& path) {
    if (!fs::exists(path)) {
        fs::create_directories(path);
    }
}

// Creates an empty file.
void TreeStructurer::create_file(const fs::path& path) {
    if (!fs::exists(path)) {
        std::ofstream ofs(path);
        if (!ofs.is_open()) {
            throw std::runtime_error("Failed to create file: " + path.string());
        }
        ofs.close();
    }
}

// Reads a structure file into a vector of non-empty lines.
std::vector<std::string> TreeStructurer::read_structure_file(const std::string& filepath) {
    std::vector<std::string> lines;
    std::ifstream file(filepath);
    if (!file.is_open()) {
        throw std::runtime_error("Failed to open file: " + filepath);
    }
    std::string line;
    while (std::getline(file, line)) {
        if (!line.empty()) {
            lines.push_back(line);
        }
    }
    return lines;
}

// Checks the structure for obvious mistakes (e.g. a jump in indentation).
void TreeStructurer::validate_structure(const std::vector<std::string>& lines) {
    if (lines.empty()) {
        throw std::runtime_error("Empty structure provided");
    }
    size_t prev_indent = 0;
    for (const auto& line : lines) {
        if (line.empty()) continue;
        size_t indent = get_indent_level(line);
        std::cout << "Line: \"" << line << "\" Indent level: " << indent << std::endl;
        if (indent > prev_indent + 1) {
            throw std::runtime_error("Invalid indentation structure in the file");
        }
        prev_indent = indent;
    }
}

// Removes any disallowed characters from a node name (here we allow printable ASCII and '/').
std::string TreeStructurer::sanitize_path(const std::string& path) {
    std::string result;
    for (char c : path) {
        if ((c >= 32 && c <= 126) || c == '/') {
            result.push_back(c);
        }
    }
    return result;
}
